from .CartesianProduct import product
from .Types import SpecialGenerator


__all__ = ["pure_bruteforce_wordlist", "_pure_bruteforce_rule"]


# chars, minlength, maxlength
def _pure_bruteforce_rule(rule: str):
    """Function that creates a virtual wordlist based on rules.

    Necessary:
        - `chars=string` : All the chars that are going to be used during the cartesian product
        - `minlength=int` : Minimum length of generated words
        - `maxlength=` : maximum length of generated words

    Arguments:
        - rule: The rule the is going to be used to create the virtual wordlist

    Returns:
        - yield each word generated by the virtual wordlist
    """
    # noinspection PyTypeChecker
    rule = dict(value.split("=") for value in rule.split(","))
    rule["minlength"] = int(rule["minlength"])
    rule["maxlength"] = int(rule["maxlength"])
    for length in range(rule["minlength"], rule["maxlength"] + 1):
        for word in product(rule["chars"], length):
            yield "".join(word)


def pure_bruteforce_wordlist(rule):
    return SpecialGenerator(_pure_bruteforce_rule, rule)



import multiprocessing
import time
import threading
import sys
import collections
import argparse

from .Arguments import Arguments
from .Process import Process
from .Modes import ARGUMENTS_MODE
from .Modes import WORDLIST_MODE


__all__ = ["Runner", "_get_arguments", "_queue_arguments_loader"]


# Default argument capture for the main function
def _get_arguments() -> Arguments:
    """Default function to prepare the arguments for the `Runner` during it's execution in a terminal

    Returns:
        - bftool.Arguments with all  the configurations provided by the user
    """
    argument_parser = argparse.ArgumentParser()
    argument_parser.add_argument("-mt", "--max-threads",
                                 help="Maximum number of threads per process", default=1, type=int)
    argument_parser.add_argument("-mp", "--max-processes",
                                 help="Maximum number of process to have active at the same time",
                                 default=ARGUMENTS_MODE, type=int)
    argument_parser.add_argument("-w", "--wordlist", help="File wordlist to use"
                                                          " based on \"[ARGUMENT_INDEX, ARGUMENT_NAME]:file_path\"",
                                 action="append", default=[])
    argument_parser.add_argument("-b", "--bruteforce",
                                 help="Generate a virtual wordlist based on \
                                 rules \"[ARGUMENT_INDEX, ARGUMENT_NAME]:chars=...,minlength=...,maxlength=...\"",
                                 action="append", default=[])
    argument_parser.add_argument("-m", "--mode",
                                 help="Mode to use during the function execution (way to divide the threads)",
                                 choices=("wordlist", "arguments"), default="arguments")
    argument_parser.add_argument("-sf", "--success-function",
                                 help="Function to pass the success result (default is 'print')",
                                 default=print)
    argument_parser.add_argument('--debug-off',
                                 help="Do not print the setup messages",
                                 default=False,
                                 action="store_true")
    argument_parser.add_argument("script_path", help="Python script to import")
    argument_parser.add_argument("function_name", help="Name of the function implemented in the python script to use")
    parsed_arguments = argument_parser.parse_args()
    if parsed_arguments.mode == "wordlist":
        parsed_arguments.mode = WORDLIST_MODE
    elif parsed_arguments.mode == "arguments":
        parsed_arguments.mode = ARGUMENTS_MODE
    arguments = Arguments(
        script_path=parsed_arguments.script_path,
        debug=not parsed_arguments.debug_off,
        function_name=parsed_arguments.function_name,
        success_function=parsed_arguments.success_function,
        wordlists_files=dict(wordlist.split(":") for wordlist in parsed_arguments.wordlist),
        wordlists_pure_bruteforce_rules=dict(wordlist.split(":") for wordlist in parsed_arguments.bruteforce),
        maximum_number_of_concurrent_processes=parsed_arguments.max_processes,
        maximum_number_of_process_threads=parsed_arguments.max_threads,
        fuzzing_mode=parsed_arguments.mode,
    )
    return arguments


def _queue_arguments_loader(arguments_queue: multiprocessing.Queue, master_wordlist: iter):
    """Fill the `arguments_queue` of the `Main` with the arguments provided by `Wordlist`


    Arguments:
        - arguments_queue: multiprocessing.Queue that is shared with active processes
        - wordlist_: bftool.Wordlist.Wordlist object
    """
    for argument in master_wordlist:
        arguments_queue.put(argument)
    exit(0)


# High level class to handle the distribution of execution of a python function
class Runner(object):
    """This class provide a high level interface for the execution distribution of a python function, it is
    the main core of the module"""

    def __init__(self):
        self.__debug_setup = True
        # Wordlist distributed processes
        self.__processes = []
        self.__print_queue = multiprocessing.Queue()
        self.__finish = False
        self.__running = False

    # Use this in win32 system as it has trouble using multiprocessing.Process().start()
    def start_win32_process(self, process: multiprocessing.Process):
        """Only a Windows friendly way to start a process

        Arguments:
            - process: process to start
        """
        process_thread = threading.Thread(target=process.run, )
        self.__processes.append(process_thread)

    # Use this function to handle prints
    def success_queue_(self, success_function: collections.abc.Callable):
        """This need to run in the background as it is the one that print all
        the results that the print_queue contains"""
        while not self.__finish:
            success_function(self.__print_queue.get())
        while not self.__print_queue.empty():
            success_function(self.__print_queue.get())

    def _print_setup(self, message: str):
        if self.__debug_setup:
            print(message)

    # Use this function to start bftool
    def run(self, arguments: Arguments = None):
        """This function activates the Main, It distributes the function in the
         specified processes and threads. To understand how it's arguments can be prepared, please check
         `bftool.ArgumentConstructor.Arguments`

         Arguments:
             - arguments: bftool.ArgumentConstructor.Arguments
        """
        if self.__running:
            if not self.__finish:
                raise RuntimeError("Can\'t use a runner that is already being used")
            self.__processes.clear()
            self.__print_queue.empty()
            self.__running = False
            self.__debug_setup = True
        self.__running = True
        total_time = time.time()
        # If the user use bftool as a module he may want to specify a custom set of arguments
        if arguments is None:
            arguments = _get_arguments()
        arguments.is_valid()
        self.__debug_setup = arguments.debug

        # Get the function to be used by the s
        function_ = arguments.function

        # Thanks to this it is possible to distribute the input arguments to multiple threads in a thread safe manner
        wordlist_queue = multiprocessing.Queue()

        # Start filling the queue with the arguments for the functions
        wordlist_queue_thread = threading.Thread(target=_queue_arguments_loader,
                                                 args=(wordlist_queue, arguments.master_wordlist)
                                                 )
        wordlist_queue_thread.start()

        self._print_setup("--- Starting child processes ---")
        # Start the print queue
        print_queue_thread = threading.Thread(target=self.success_queue_, args=(arguments.success_function,))
        print_queue_thread.start()

        processes_setup_time = time.time()

        # Preparing all child processes
        for index in range(1, arguments.maximum_number_of_concurrent_processes + 1):
            process_ = Process(index,
                               function_,
                               wordlist_queue,
                               arguments.maximum_number_of_process_threads,
                               arguments.fuzzing_mode,
                               self.__print_queue)
            if sys.platform == "win32":
                self.start_win32_process(process_)
            else:
                self.__processes.append(process_)

            self._print_setup(f"* Process with ID {index} - Prepared")
        processes_setup_time = time.time() - processes_setup_time
        self._print_setup("--- All processes were prepared ---")
        self._print_setup("--- Waiting to finish ---")
        fuzzing_time = time.time()
        # Starting all the processes
        for process in self.__processes:
            process.start()
        # Waiting all processes to finish
        for process_ in self.__processes:
            process_.join()
        fuzzing_time = time.time() - fuzzing_time
        wordlist_queue_thread.join()
        total_time = time.time() - total_time
        self._print_setup("--- END ---")
        self._print_setup(f"Time setting up the processes: {processes_setup_time}")
        self._print_setup(f"Time fuzzing: {fuzzing_time}")
        self._print_setup(f"Total time: {total_time}")
        # Kill switch
        self.__finish = True
        print_queue_thread.join()
        exit(0)



import collections
import inspect
import string

from .Modes import ARGUMENTS_MODE
from .ImportCode import import_function_from_script
from .CartesianProduct import combine_wordlists
from .BruteforceWordlist import pure_bruteforce_wordlist
from .WordlistFromFile import read_file_lines


__all__ = ["Arguments"]


# This class prepare all the arguments that the Runner is going to use
class Arguments(object):
    """Class that handles all the configuration of the the arguments that are going to be passed
    to `bftool.Runner`"""

    def __init__(self,
                 function_: collections.abc.Callable = None,
                 success_function: [collections.abc.Callable, str] = print,
                 debug: bool = True,
                 script_path: str = None,
                 function_name: str = None,
                 wordlists_iterables: dict = None,
                 wordlists_pure_bruteforce_rules: dict = None,
                 wordlists_files: dict = None,
                 maximum_number_of_concurrent_processes: int = 1,
                 maximum_number_of_process_threads: int = 1,
                 fuzzing_mode=ARGUMENTS_MODE,
                 ):
        """
            - success_function: The function to use to pass the results of the executed function
            - debug: print the setup messages
            - function_: The function object to be used
            - script_path: The script path with the source code of the wanted function (Incompatible with `function_`)
            - function_name: The function to be imported (Incompatible with `function_`)
            - wordlists_iterables: Dictionary with the iterables that are going to be passed to the arguments, please
            - notice that if you want to pass a generator object you may construct it with bftool.Types.SpecialGenerator
            - wordlists_pure_bruteforce_rules: Dict with the bruteforce rules that are going to be passed to the
            function
            - wordlists_files: Dict with the wordlists paths that are going to be passed to the function
            - maximum_number_of_concurrent_processes: maximum number of concurrent processes
            - maximum_number_of_process_threads: maximum number of concurrent threads per process
            - fuzzing_mode: the fuzzing mode
        """
        self._load_success_function(success_function, script_path)

        self.debug = debug

        self._load_function(function_, script_path, function_name)

        # - Wordlist Setup
        # This is the argument that will replace replace Wordlist.wordlist
        self.__wordlists = [None] * len(self.__function_args_spec.args)

        self._load_words_from_iterables(wordlists_iterables)
        self._load_words_from_bruteforce_rules(wordlists_pure_bruteforce_rules)
        self._load_words_from_files(wordlists_files)

        self.master_wordlist = combine_wordlists(self.__wordlists)

        self.maximum_number_of_concurrent_processes = maximum_number_of_concurrent_processes
        self.maximum_number_of_process_threads = maximum_number_of_process_threads
        self.fuzzing_mode = fuzzing_mode

    def _load_success_function(self, success_function: [collections.abc.Callable, str], script_path):
        if isinstance(success_function, collections.abc.Callable):
            self.success_function = success_function
        elif isinstance(success_function, str) and isinstance(script_path, str):
            self.success_function = import_function_from_script(script_path, success_function)
        else:
            raise TypeError("Success function must be callable")

    def _load_function(self,
                       function_: collections.abc.Callable = None,
                       script_path: str = None,
                       function_name: str = None):
        # - Function Setup
        # When the users bftool as a module and not a script he can provide the raw function object
        if isinstance(function_, collections.abc.Callable):
            self.function = function_
        # When the user specify that the function is going to be extracted of another python file
        elif isinstance(script_path, str) and isinstance(function_name, str):
            self.function = import_function_from_script(script_path, function_name)
        else:
            raise ValueError("No function provided")
        self.__function_args_spec = inspect.getfullargspec(self.function)

    def _load_words_from_iterables(self, iterables_wordlists: dict):
        # If the user provide raw iterables like list, sets, tuples, dicts...
        # To handle generators please check bftool.Types.SpecialGenerator
        if iterables_wordlists:
            for key, value in iterables_wordlists.items():
                if isinstance(key, int):
                    self.__wordlists[key] = value
                elif isinstance(key, str):
                    if key[0] in string.digits:
                        self.__wordlists[int(key)] = value
                    else:
                        self.__wordlists[self.__function_args_spec.args.index(key)] = value
                else:
                    raise KeyError(f"Can't index in the function arguments key {key} of type {type(key)}")

    def _load_words_from_bruteforce_rules(self, bruteforce_rules_wordlists: dict):
        # If the user provides wordlist generation rules
        if bruteforce_rules_wordlists:
            for key, value in bruteforce_rules_wordlists.items():
                if isinstance(key, int):
                    self.__wordlists[key] = pure_bruteforce_wordlist(value)
                elif isinstance(key, str):
                    if key[0] in string.digits:
                        self.__wordlists[int(key)] = pure_bruteforce_wordlist(value)
                    else:
                        self.__wordlists[self.__function_args_spec.args.index(key)] = pure_bruteforce_wordlist(value)
                else:
                    raise KeyError(f"Can't index in the function arguments key {key} of type {type(key)}")

    def _load_words_from_files(self, files_wordlists: dict):
        # If the user provide a file as a wordlist (it always read each line as an argument)
        if files_wordlists:
            for key, value in files_wordlists.items():
                if isinstance(key, int):
                    self.__wordlists[key] = read_file_lines(value)
                elif isinstance(key, str):
                    if key[0] in string.digits:
                        self.__wordlists[int(key)] = read_file_lines(value)
                    else:
                        self.__wordlists[self.__function_args_spec.args.index(key)] = read_file_lines(value)
                else:
                    raise KeyError(f"Can't index in the function arguments key {key} of type {type(key)}")

    def is_valid(self):
        """
        This functions checks that the prepared arguments are valid
        :return:
        """
        if not self.__wordlists:
            raise ValueError("No wordlist input provided")

        if any(value is None for value in self.__wordlists):
            raise IndexError(f"Invalid number of wordlists provided ({len(self.__wordlists)})"
                             f"for a function of {len(self.__function_args_spec.args)} arguments")
        return True



import threading
import types
import queue
import multiprocessing


__all__ = ["Thread"]


# Only a way to handle threads inside processes
class Thread(threading.Thread):
    """Independent Thread, that did not spawn a independent thread for each function execution as it is in self the
    thread that calls it, in other words, it is the equivalent of `bftool.Process` in arguments mode but without
    spawning an independent thread for each function call"""
    def __init__(self, wordlist_queue: multiprocessing.Queue,
                 function_: types.FunctionType, print_queue: multiprocessing.Queue):
        """
            - wordlist_queue: the master wordlist queue
            - function_: the function that is going to be executed
            - print_queue: the printing queue to put the results
        """
        # Initialize the original thread elements
        threading.Thread.__init__(self)
        self.__wordlist_queue = wordlist_queue
        self.__function = function_
        self.__print_queue = print_queue

    # Thread runner
    def run(self):
        """This function is executed when the user calls `Thread.start()`"""
        while True:
            try:
                arguments = self.__wordlist_queue.get(True, timeout=5)
            except queue.Empty:
                break
            result = self.__function(*arguments)
            if result is not None:
                self.__print_queue.put(result)
        exit(0)



import collections


__all__ = ["SpecialGenerator", "ExpandableTuple"]


# Low cost way to backup python generators for re-iterations
class SpecialGenerator(object):
    """Low cost way to backup python generators for re-iterations, so always use it when you whant to pass generators to
    `bftool.Arguments` so the arguments distribution generator (the master one) can understand how to revert it to it's
    original state"""
    def __init__(self, generator_function: collections.abc.Callable, *args):
        """
        Arguments:
            - generator_function: The function that creates the generator
            - args: the arguments to pass to that function
        """
        self.function = generator_function
        self.arguments = args

    def __call__(self, *args, **kwargs):
        return self.function(*self.arguments)


# Easies way to not expand arguments that are tuple
class ExpandableTuple(tuple):
    """Easies way to not expand arguments that are tuple, so if you are not going to work with the internal of the
    module please leave it"""
    def __str__(self):
        return f"ExpandableTuple({', '.join(str(value) for value in self)})"

    def __repr__(self):
        return f"ExpandableTuple({', '.join(str(value) for value in self)})"



import os
import importlib
import types


__all__ = ["import_module_from_path", "import_function_from_script"]


# Import a python module from it's filesystem path
def import_module_from_path(main_py_path: str) -> types.ModuleType:
    """Import a python module from it's FileSystem path

    Arguments:
        - main_py_path: path to the python script to import

    Returns:
        - module object
    """
    if not os.path.exists(main_py_path):
        raise FileExistsError("Looks like file not exists")
    if not os.path.isfile(main_py_path):
        raise TypeError("Looks like path supplied is not a file")
    module_name = main_py_path.replace("\\", "/").split("/")[-1].split(".")[0]
    return importlib.import_module(module_name, main_py_path)


def import_function_from_script(script_path: str, function_name: str):
    """Function to extract the a python function from a module

    Arguments:
        - script_path: path to the python script
        - function_name: name of the function to import

    Returns:
        - Returns the object named by function_name
    """
    module = import_module_from_path(script_path)
    return getattr(module, function_name)



import threading
import types
import multiprocessing
import queue

from .Thread import Thread

from .Modes import ARGUMENTS_MODE
from .Modes import WORDLIST_MODE


__all__ = ["Process"]


# This class is a high level API implemented by inheritance with the python multiprocessing.Process
class Process(multiprocessing.Process):
    """This class provide a high level API to handle the distribution of threads inside a independent process"""
    def __init__(self,
                 process_id: int,
                 function_: types.FunctionType, wordlist_queue: multiprocessing.Queue,
                 max_threads: int, mode: int,
                 print_queue):
        """
            - process_id: ID of the process
            - function_: function that is goning to be passed to the threads
            - wordlist_queue: the master wordlist queue
            - max_threads: The maximum number of concurrent threads for this process
            - mode: The mode in which distribute the execution
            - print_queue: master print queue
        """
        multiprocessing.Process.__init__(self)
        self.__process_id = process_id
        self.__function = function_
        self.__wordlist_queue = wordlist_queue
        self.__active_threads = 0
        self.__max_threads = max_threads
        self.__mode = mode
        self.__threads_queue = queue.Queue()
        self.__print_queue = print_queue

    #  for the wordlist mode
    def _wordlist_block_mode(self):
        """This is used when the user specify the mode WORDLIST, it spawns independent threads that handle function"""
        for index in range(self.__max_threads):
            if self.__active_threads > self.__max_threads:
                self.__threads_queue.get().join()
                self.__active_threads -= 1
            wordlist_block_thread = Thread(self.__wordlist_queue,
                                           self.__function, self.__print_queue)
            self.__threads_queue.put(wordlist_block_thread)
            wordlist_block_thread.start()
            self.__active_threads += 1
        while self.__active_threads > 0:
            self.__threads_queue.get().join()
            self.__active_threads -= 1

    #  for the execution of the the specified function in a independent thread (ARGUMENTS mode)
    def __arguments_mode_function(self, *args):
        """ for the execution of the the specified function in a independent thread (`ARGUMENTS` mode)

        Arguments:
            - args: arguments that are going to be passed to the function
        """
        result = self.__function(*args)
        if result is not None:
            self.__print_queue.put(result)
        exit(0)

    #  for the arguments mode, it spawn an independent thread for each function execution
    def _arguments_block_mode(self):
        """ for the arguments mode, it spawn an independent thread for each function execution"""
        while True:
            try:
                arguments = self.__wordlist_queue.get(True, timeout=5)
            except queue.Empty:
                break
            if self.__active_threads >= self.__max_threads:
                # print(self.__active_threads)
                self.__threads_queue.get().join()
                self.__active_threads -= 1
            arguments_thread = threading.Thread(target=self.__arguments_mode_function, args=arguments)
            self.__threads_queue.put(arguments_thread)
            arguments_thread.start()
            self.__active_threads += 1
        while self.__active_threads > 0:
            self.__threads_queue.get().join()
            self.__active_threads -= 1

    # Runner, it start the process
    def run(self):
        """Runner, it start the process, if you are in Linux please consider to use `Process.start()`, if you are
        in Windows consider tu execute this function in a independent thread with the threading module, it could look
        like this `threading.Thread(target=Process.run, args=(*args, **kargs)).start()`"""
        if self.__mode == WORDLIST_MODE:
            self._wordlist_block_mode()
        elif self.__mode == ARGUMENTS_MODE:
            self._arguments_block_mode()
        else:
            raise KeyError("Unknown mode supplied")
        self.__print_queue.put(f"* Process {self.__process_id} - Done")
        exit(0)



from .Types import SpecialGenerator
from .Types import ExpandableTuple


__all__ = ["product", "expand_product_result", "combine_wordlists"]


def product(iterable_: iter, length: int, join=False):
    """Custom cartesian product that creates product on the fly

    Arguments:
        - iterable_: iterable object
        - length: length of products generated
        - join: (Default=False) When the iterable is a string you can put this to True so the generated words are string

    Returns:
        - Yields a tuple if join is set to False otherwise yields strings
    """
    try:
        iterable_ = tuple(set(iterable_))
    except TypeError:
        pass
    if length < 1:
        raise ValueError("Invalid length supplied")
    indexes = [0] * length
    chars_length = len(iterable_)
    while True:
        if not join:
            yield tuple(iterable_[index] for index in indexes)
        else:
            yield "".join(iterable_[index] for index in indexes)
        for index, index_value in enumerate(indexes):
            if indexes[index] < chars_length - 1:
                indexes[index] += 1
                break
            else:
                indexes[index] = 0
        if all(index == 0 for index in indexes):
            break


def expand_product_result(product_: tuple) -> list:
    """This function expand in it's correct way, the arguments provided by custom product

    Arguments:
        - product: The tuple containing the arguments

    Returns:
        - A sanitized tuple of arguments (sanitized as it eliminates one tuple in another, in another ...)
    """
    result = []
    for value in product_:
        if isinstance(value, ExpandableTuple):
            result += expand_product_result(value)
        elif isinstance(value, tuple):
            if len(value) == 1:  # If the tuple is just one argument
                result.append(value[0])
            else:
                result.append(value)
        else:
            result.append(value)
    return result


# Maybe useless
# Expected input
# [bftool.Types.SpecialGenerator, list, tuple, set, dict, ...] (list of iterables)
def combine_wordlists(iterables_: list, master=True):
    """This function is intended to handle the cartesian product of all arguments

    Arguments:
        - iterables_: This should be a list contains all the bruteforce rules, file reader generator and iterables you
        - want to combine

    Returns:
        - It yields the combinations of the input wordlists
    """
    number_of_iterables = len(iterables_)
    if isinstance(iterables_[0], SpecialGenerator):
        cycle_iterable = iterables_[0]()
    else:
        cycle_iterable = iterables_[0]
    if number_of_iterables == 1:
        for value in cycle_iterable:
            yield value,
    elif number_of_iterables > 1:
        for value in cycle_iterable:
            for second_value in combine_wordlists(iterables_[1:], False):
                second_value = ExpandableTuple(second_value)
                if master:
                    # When is the master function (the one that the user called), normalize it's values
                    yield expand_product_result((value, second_value))
                else:
                    yield value, second_value
    else:
        raise IndexError("Invalid number of arguments")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bftool-pkg-sulcud",
    version="1.4.5.1",
    author="Antonio Donis",
    author_email="antoniojosedonishung@gmail.com",
    description="A python module and script to handle custom fuzzing against anything",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/shoriwe/bftool",
    packages=["bftool"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from .Types import SpecialGenerator


__all__ = ["_read_file_lines", "read_file_lines"]


def _read_file_lines(file_path: str):
    """Function that iterates over a file lines, eat less memory than reading it at once

    Arguments:
        - file_path: Path to the wordlist file

    Returns:
        - Yield each line of the file
    """
    file_object = open(file_path, errors="ignore")
    for line in file_object:
        yield line[:-1] if line[-1] == "\n" else line
    file_object.close()


def read_file_lines(file_path: str) -> SpecialGenerator:
    return SpecialGenerator(_read_file_lines, file_path)
